@@ -26,6 +26,9 @@ radeon_legacy_allocate_memory(ScrnInfoPtr pScrn,
     if (info->cs) {
 	struct radeon_bo *video_bo;
 
+	if (*mem_struct)
+		radeon_legacy_free_memory(pScrn, *mem_struct);
+
 	video_bo = radeon_bo_open(info->bufmgr, 0, size, RADEON_GPU_PAGE_SIZE, domain, 0);
 
 	*mem_struct = video_bo;
